#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Module that contain base class for read data


from shutil import which
from ..exceptions.exception import HandlerException
from ..yamloperations.yamloperations import YAMLOperations
from ..utilities.extcommand import run_external_cmd


# Base class for read data from the chip or simulator
class NICExchangeData:
    def __init__(self, ibv_device, debug):
        self._ibv_device = ibv_device
        self._vhca_data = None
        self._resource_data = {}
        self._cmd_list = []
        self._debug = debug
        self._version = None
        self._execname = None
        self._sudo_prefix = None

    # Function for check if the program exist.
    def _check_if_program_exist(self, progname):
        path = which(progname)
        if not path:
            raise HandlerException(f"The program {progname} does not exist")

    # Function to create a list of parameters for calling a utility to
    # read information.
    def _construct_info_list(self, sub_cmd):
        self._cmd_list = [self._get_sudo_prefix(sub_cmd), self._get_info_exec(sub_cmd)]
        self._cmd_list.extend([sub_cmd])
        self._cmd_list.extend(["--dpa_device", self._ibv_device])

    # Function to create a list of parameters for calling a utility to
    # read information.
    def _construct_info_list(self, sub_cmd):
        self._cmd_list = [self._get_sudo_prefix(sub_cmd), self._get_info_exec(sub_cmd)]
        self._cmd_list.extend([sub_cmd])
        self._cmd_list.extend(["--dpa_device", self._ibv_device])

    # Function to create a list of parameters for calling a utility to
    # create, destroy, query partitions and groups.
    def _construct_cmd_list(self, cmd, sub_cmd, kwarg):
        self._cmd_list = [self._get_sudo_prefix(cmd, sub_cmd), self._get_cmd_exec(sub_cmd)]
        self._cmd_list.extend([cmd, sub_cmd])
        self._cmd_list.extend(["--dpa_device", self._ibv_device])
        for key, value in kwarg.items():
            if len(key) == 1:
                self._cmd_list.extend(['-' + key, str(value)])
            else:
                self._cmd_list.extend(['--' + key, str(value)])

    # Protected function to run external utility, get output, print them and
    # the command line in case of the error and return string of the output
    def _run_command(self, cmd):
        if self._debug:
            print((" ".join(cmd)).strip())
        ret, out, err = run_external_cmd(cmd)
        if ret:
            print("Command is failed")
            if not self._debug:
                print((" ".join(cmd)).strip())
            if err and len(err):
                print("".join(err))
            if out and len(out):
                print("".join(out))
            raise HandlerException(f"Running failed with error ({ret}) on cmd")
        return "".join(out)

    # Protected function to run external utility for info commands
    def _read_info_data(self, sub_cmd):
        self._construct_info_list(sub_cmd)
        return self._run_command(self._cmd_list)

    # Protected function for parse partitions capabilities
    def _parse_partition_resources(self, output):
        input_keys = {"Max number of DPA EU partitions": "max_num_dpa_eu_partitions",
                      "Max number of VHCAs associated": "max_num_partition_vhca_id",
                      "Max number of DPA EU groups": "max_num_dpa_groups"}

        out_list = output.split("\n")
        for out_string in out_list:
            for key, resource_key in input_keys.items():
                if key in out_string:
                    self._resource_data[resource_key] = int(out_string.split(":")[1].strip())

    # Protected function for parse groups capabilities
    def _parse_groups_resources(self, output):
        input_keys = {"Max number of DPA EUs in one DPA EU group": "max_num_dpa_eu_per_group",
                      "Max DPA EU number available to use": "max_num_dpa_eu"}
        out_list = output.split("\n")
        for out_string in out_list:
            for key, resource_key in input_keys.items():
                if key in out_string:
                    self._resource_data[resource_key] = int(out_string.split(":")[1].strip())

    # Public function to run external utility for create, destroy, query commands
    def run_command(self, cmd, sub_cmd, **kwarg):
        self._construct_cmd_list(cmd, sub_cmd, kwarg)
        output = self._run_command(self._cmd_list)
        return output.split('\n')

    # Protected function to get command name
    def _get_cmd_exec(self, *_):
        return self._execname

    # Protected function to get sudo prefix
    def _get_info_exec(self, _):
        return self._execname

    # Protected function to get sudo prefix
    def _get_sudo_prefix(self, *_):
        return self._sudo_prefix

    # Public function to read vhca data
    def read_vhca_data(self):
        output = self._read_info_data("info vhca")
        self._vhca_data = YAMLOperations().get_object_from_yaml(output)

    # Public function to read resources
    def read_dpa_resources(self):
        output = self._read_info_data("info chip")
        self._resource_data.update(YAMLOperations().get_object_from_yaml(output))

    # Public function to read partitions resources
    def read_dpa_partitions_resources(self):
        output = self._read_info_data("partition info")
        self._parse_partition_resources(output)

    # Public function to read groups resources
    def read_dpa_groups_resources(self):
        output = self._read_info_data("eu_group info")
        self._parse_groups_resources(output)

    # Public function to get the version of the tool
    def get_tool_version(self):
        cmd = "version"
        command_line = [self._get_sudo_prefix(cmd), self._get_cmd_exec(cmd), cmd]
        data = self._run_command(command_line)
        self._version = ".".join(data.replace("Version: ", "").replace("'", "").
                                 strip().split('.')[:2])
        return self._version

    # Public function to read partitions query data and return it as a list of strings
    def read_dpa_partitions_query(self):
        return self.run_command("partition", "query")

    # Public function to read groups query data and return it as a list of strings
    def read_dpa_groups_query(self, partition_id=None):
        if partition_id:
            output = self.run_command("eu_group", "query", id_partition=partition_id)
        else:
            output = self.run_command("eu_group", "query")
        return output

    # Public function to read vhca query data and return it as a YAML
    @property
    def vhca_data(self):
        return self._vhca_data

    # Public function to read resources data and return it as a dictionary
    @property
    def resource_data(self):
        return self._resource_data

    # Public function to get status of chip it as a dictionary - how many partitions,
    # root groups and processes created
    def get_status(self):
        output = self._read_info_data("info status")
        return YAMLOperations().get_object_from_yaml(output)
