#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from ..exceptions.exception import HandlerException
from ..yamloperations.yamloperations import YAMLOperations
from ..utilities.string import create_string_from_list

RESOURCE_KEYS = [
    'max_eu_groups',
    'max_num_dpa_eu_partitions',
    'max_num_dpa_eu_per_group',
    'max_num_partition_vhca_id',
    'max_num_dpa_eu',
    'num_cores',
    'cores',
]

RESOURCE_DICTIONARY = {
    "max_num_dpa_eu_per_group": "Maximal number of DPA EU per group",
    "max_num_dpa_eu": "Maximal number of active DPA EU",
    "max_num_dpa_eu_partitions": "Maximal number of DPA EU partitions",
    "max_num_partition_vhca_id": "Maximal number of VHCAs associated with a single partition",
    "max_num_dpa_groups": "Maximal number of EU groups could be created",
    "heap_max_size": "Heap maximal size",
    "max_eu_groups": "Maximal number of EU groups by FirmWare",
    "max_process_per_vhca": "Maximal number of processes per VHCA",
    "max_processes": "Maximal number of DPA processes",
    "max_threads": "Maximal number of DPA threads",
    "max_threads_per_process": "Maximal number of DPA threads per process",
    "num_cores": "Number of cores",
    "thread_stack_size": "DPA thread stack size",
}


# Class of resources
class DPAResources:
    def __init__(self, instant_data):
        self._data = instant_data
        self._check_and_add_keys(instant_data)

    # Function for chack that all keys are exist
    def _check_and_add_keys(self, instant_data):
        missing_keys = [key for key in RESOURCE_KEYS if key not in instant_data]

        if missing_keys:
            raise HandlerException(f"Keys {missing_keys} are missing "
                                   "in DPA resources")

        for key, value in instant_data.items():
            setattr(self, '_' + key, value)

    # Function for write data as yaml
    def write_dict_as_yaml(self, output):
        for key in RESOURCE_DICTIONARY:
            if key in self._data:
                output.write(f"{key}: {self._data[key]}\n")
                del self._data[key]
        YAMLOperations().write_dict(self._data, output)

    # Function for print resources
    def print_resources(self, output):
        for key, value in RESOURCE_DICTIONARY.items():
            if key in self._data:
                output.write(f"{value} ({key}): {self._data[key]}\n")
        output.write("Cores:\n")
        for core in self._cores:
            is_last = core == self._cores[-1]
            spacing = " " * 4
            connector = "└── " if is_last else "├── "
            space = "    " if is_last else "│   "
            output.write(f"{spacing}{connector}Core ID {core['core_id']}\n")
            output.write(f"{spacing}{space}EUs number {core['eu_num']}\n")
            list_eu_ids = [eu['eu_id'] for eu in core['eus']]
            list_eu_indexes = [eu['eu_index'] for eu in core['eus']]
            output.write(f"{spacing}{space}EUs IDs {create_string_from_list(list_eu_ids)}\n")
            output.write(f"{spacing}{space}EUs indexes "
                         f"{create_string_from_list(list_eu_indexes)}\n")
            output.write(f"{spacing}{space}\n")

    @property
    def max_num_partitions(self):
        return self._max_num_dpa_eu_partitions

    @property
    def max_eu_groups(self):
        return self._max_eu_groups

    @property
    def max_eu_per_group(self):
        return self._max_num_dpa_eu_per_group

    @property
    def max_vhca_id_size(self):
        return self._max_num_partition_vhca_id

    @property
    def max_num_dpa_eu(self):
        return self._max_num_dpa_eu

    @property
    def num_cores(self):
        return self._num_cores

    @property
    def cores(self):
        return self._cores
