#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from ..exceptions.exception import HandlerException
TREE_DEPTH = 10


# Main class for VHCA Tree and Array data
class VHCATreeData:
    def __init__(self, instant_data):
        self._obj_data = {}
        self._obj_dict = {}
        self._obj_tree = {}
        self._create_data(instant_data)

    @property
    def obj_dict(self):
        return self._obj_dict

    @property
    def obj_tree(self):
        return self._obj_tree

    # Protected method for get id name
    def _get_id_name(self):
        return 'vhca_id'

    # Protected method for get parent id name
    def _get_parent_id_name(self):
        return 'parent_vhca_id'

    # Protected method for print fields that specific for class
    def _print_specific_fields(self, elem):
        return f"{elem['function_type']}"

    # Protected method for set fields that specific for class
    def _set_dict_specific_fields(self, elem, instant_elem):
        specific_list = ["function_type", "pci_bdf", "vport_id", "parent_vhca_id"]
        for key in specific_list:
            if key in instant_elem:
                elem[key] = instant_elem[key]

    # Protected method for create node
    def _create_node(self, instant_elem):
        elem = {}
        id_name = self._get_id_name()
        parent_id_name = self._get_parent_id_name()
        id_val = instant_elem[id_name]
        elem[id_name] = instant_elem[id_name]
        self._set_dict_specific_fields(elem, instant_elem)
        elem['children'] = []
        if parent_id_name in instant_elem:
            parent_id = instant_elem[parent_id_name]
            if parent_id not in self._obj_dict:
                return False
            parent = self.obj_dict[parent_id]
            elem['parent'] = parent
            parent['children'].append(elem)
        else:
            elem['parent'] = None
            if 'root' not in self._obj_tree:
                self._obj_tree['root'] = []
            self._obj_tree['root'].append(elem)
        self._obj_dict[id_val] = elem
        return True

    # Protected method for create tree
    # This function create tree from the array of functions. Because
    # the array is not sorted it done by the loop. All elements of
    # teh array should be handled and if not then will be exception.
    def _create_data(self, instant_data):
        self._obj_data = instant_data
        size_of_array = len(instant_data)
        handling_array = [0] * size_of_array
        all_handled = False

        for _ in range(0, TREE_DEPTH):
            handled = 0
            for index in range(0, size_of_array):
                if handling_array[index] == 0:
                    if self._create_node(instant_data[index]):
                        handling_array[index] = 1
                        handled += 1
                else:
                    handled += 1
            if handled == size_of_array:
                all_handled = True
                break

        if not all_handled:
            raise HandlerException("Cannot handle vhca array to tree")

    def get_elem(self, key):
        if key in self.obj_dict:
            return self.obj_dict[key]
        return None

    def get_key(self, elem):
        if not elem:
            return None
        try:
            return int(elem[self._get_id_name()])
        except KeyError:
            return None

    def get_parent(self, elem):
        if not elem:
            return None
        return elem['parent']

    def print_sub_tree(self, output, elem, offset=0, prefix="", is_last=True):
        func_list = ["PF", "VF", "SF", "ECPF"]

        if isinstance(elem, list):
            for i, child_elem in enumerate(elem):
                self.print_sub_tree(output, child_elem, offset, prefix, i == (len(elem) - 1))
            return

        spacing = " " * 4
        connector = "└── " if is_last else "├── "
        space = "    " if is_last else "│   "
        new_prefix = prefix + space + spacing

        output.write(f"{prefix}{connector}VHCA_ID = {elem['vhca_id']:#x}\n")
        output.write(f"{prefix}{space}FUNCTION_TYPE = {func_list[elem['function_type']]}\n")
        if 'pci_bdf' in elem:
            output.write(f"{prefix}{space}PCI_BDF = {elem['pci_bdf']}\n")
        if 'vport_id' in elem:
            output.write(f"{prefix}{space}VPORT_ID = {elem['vport_id']:#x}\n")
        if len(elem['children']):
            output.write(f"{prefix}{space}{spacing}│\n")
        else:
            output.write(f"{prefix}{space}\n")
        for i, child_elem in enumerate(elem['children']):
            self.print_sub_tree(output, child_elem, offset + 4, new_prefix,
                                i == (len(elem['children']) - 1))

    def print_tree(self, output, key=None):
        if not key:
            self.print_sub_tree(output, self._obj_tree['root'], 0)
        else:
            self.print_sub_tree(output, self._obj_dict[key], 0)

    def print_dict_as_yaml(self, output):
        hex_keys = ["vhca_id", "parent_vhca_id", "vport_id"]
        str_keys = ["pci_bdf"]
        for elem in self._obj_data:
            prefix = "- "
            for key, value in elem.items():
                if key in hex_keys and value:
                    string_value = f"{value:#x}"
                elif key in str_keys:
                    string_value = f"'{value}'"
                else:
                    string_value = f"{value}"
                output.write(f"{prefix}{key}: {string_value}\n")
                prefix = "  "
