#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Base module for check input database, input from the chip, create partitions
# and group on the chip, and create output yaml files.

from ..exceptions.exception import HandlerException
from ..utilities.string import expand_range_from_string, convert_values_to_list, \
    get_value_as_list, create_substring_in_range, \
    create_string_from_list, get_value_from_list_of_strings, INDEX_INVALID_VALUE
from ..constants.constants import ABS_EUS, EU_CORE, NUM_EUS, EU_ALLOCATION, EU_GROUP_ALLOCATION, \
    NAME, EUS, PARTITIONS, AFFINITY_EUS, DPA_APPS, PARTITION, ROOT_PARTITION_GROUPS, \
    ROOT_PARTITION_NAME, VHCA_ID, RELATIVE_EUS, VERSION, AFFINITY_CORE
from ..databases.dparesource import DPAResources
from ..databases.vhcatree import VHCATreeData

# Constants
EUG_MAXIMAL_LENGTH = 15


# Main class for handling
class BaseHandler:
    def __init__(self, **kwargs):
        self._main_db = {}
        self._partition_db = {}
        self._eu_map = {}
        self._eug_db = {}
        self._apps_db = {}
        self._options = kwargs['parse_options']
        self._yaml_obj = kwargs['yaml_data']
        self._chip_data = kwargs['exchange_obj'](self._options.device, debug=self._options.verbose)
        self._vhca_data = None
        self._res_data = {}

    # Create element for eu_map dictionary
    def _dict_for_eu(self, key, eu):
        return {
            'partition_name': key,
            'eu': eu,
            'eug': "",
            'rel_idx': eu
        }

    # Function for fill EU database with absolute EU
    def _fill_eu_map_with_abs(self, elem, key):
        try:
            eu_list = expand_range_from_string(str(elem[ABS_EUS]), suppress_value_error_exc=False)
        except ValueError as exc:
            raise HandlerException(f"Invalid value '{str(elem[ABS_EUS])}' of {ABS_EUS} "
                                   f"in partition {key} - {exc}") from exc
        for eu in eu_list:
            eu_map_keys = self._eu_map.keys()
            max_num_dpa_eu = self._res_data.max_num_dpa_eu
            if int(eu) >= max_num_dpa_eu:
                raise HandlerException(f"EU {eu} for {key} exceeds maximal "
                                       f"number of EU {max_num_dpa_eu}")
            if eu in eu_map_keys:
                raise HandlerException(f"EU {eu} of partition {key} is already filled")
            self._eu_map[eu] = self._dict_for_eu(key, eu)

    # Function for check existing core and eus values in the subdictionary
    def _check_core_eus(self, elem, key, eug_name=None, app_name=None):
        if app_name:
            part_string = f"in partition {key} for application {app_name}"
        else:
            part_string = f"for partition {key}"
            if eug_name:
                part_string = f"in group {eug_name} " + part_string
        if EU_CORE not in elem:
            raise HandlerException(f"Absent key '{EU_CORE}' {part_string}.")
        if NUM_EUS not in elem:
            raise HandlerException(f"Absent key '{NUM_EUS}' {part_string}.")
        if not isinstance(elem[EU_CORE], int):
            raise HandlerException(f"{EU_CORE} value {elem[EU_CORE]} {part_string} "
                                   "is not an integer")
        if not isinstance(elem[NUM_EUS], int):
            raise HandlerException(f"{NUM_EUS} value {elem[NUM_EUS]} {part_string} "
                                   "is not an integer.")
        if elem[NUM_EUS] <= 0:
            raise HandlerException(f"Invalid value {elem[NUM_EUS]} for '{NUM_EUS}' {part_string}.")

        return elem[EU_CORE], elem[NUM_EUS]

    # Function to fill EU map for partitions
    def _fill_eu_map_for_eu(self, elem, key):
        if ABS_EUS in elem:
            self._fill_eu_map_with_abs(elem, key)
            return
        core, eus = self._check_core_eus(elem, key)

        num_eus_for_core = None
        core_block = None
        for core_block in self._res_data.cores:
            if core_block['core_id'] == core:
                num_eus_for_core = core_block['eu_num']
                break
        if not num_eus_for_core:
            raise HandlerException(f"Core id {core} in {key} does not exist in DPA resources")
        if eus > num_eus_for_core:
            raise HandlerException(f"EU number {eus} in {key} exceeds "
                                   f"number of EU {num_eus_for_core}")
        start_subeu = 0
        subeu = 0
        while subeu < eus:
            if start_subeu + subeu > num_eus_for_core:
                raise HandlerException("Number of eus more than "
                                       f"{num_eus_for_core} in {key}")
            eu = core_block['eus'][start_subeu + subeu]['eu_id']
            if eu in self._eu_map:
                start_subeu += 1
                continue
            self._eu_map[eu] = self._dict_for_eu(key, eu)
            subeu += 1

    # Fill EU database with absolute EU
    def _fill_eu_map_for_eug_with_abs(self, elem, key, eug_name):
        try:
            eu_list = expand_range_from_string(str(elem[ABS_EUS]), suppress_value_error_exc=False)
        except ValueError as exc:
            raise HandlerException(f"Invalid value '{str(elem[ABS_EUS])}' of {ABS_EUS} "
                                   f"for group {eug_name} in partition {key} - {exc}") from exc
        for eu in eu_list:
            eu_map_keys = self._eu_map.keys()
            max_num_dpa_eu = self._res_data.max_num_dpa_eu
            if int(eu) >= max_num_dpa_eu:
                raise HandlerException(f"EU {eu} of {eug_name} for {key} exceeds maximal "
                                       f"number of EU {max_num_dpa_eu}")
            if key != ROOT_PARTITION_NAME:
                if eu not in eu_map_keys:
                    raise HandlerException(f"EU {eu} of group {eug_name} "
                                           f"in partition {key} is not allocated.")
            else:
                if eu in eu_map_keys:
                    raise HandlerException(f"EU {eu} of {eug_name} for {key} "
                                           "is already allocated.")
                self._eu_map[eu] = self._dict_for_eu(ROOT_PARTITION_NAME, eu)

            part_key = self._eu_map[eu]['partition_name']
            if part_key != key:
                raise HandlerException(f"EU {eu} of {eug_name} for {key} "
                                       f"allocated for partition {part_key}")
            eug_key = self._eu_map[eu]['eug']
            if eug_key != "":
                raise HandlerException(f"EU {eu} of {eug_name} for {key} "
                                       f"allocated for group {eug_key}")
            self._eu_map[eu]['eug'] = eug_name
            eug_key = (key, eug_name)
            if eug_key not in self._eug_db:
                self._eug_db[eug_key] = {}

    # Function to fill EU map for EU group
    def _fill_eu_map_for_eug(self, elem, key, eug_name):
        if ABS_EUS in elem:
            self._fill_eu_map_for_eug_with_abs(elem, key, eug_name)
            return

        core, eus = self._check_core_eus(elem, key, eug_name=eug_name)

        num_eus_for_core = None
        core_block = None
        for core_block in self._res_data.cores:
            if core_block['core_id'] == core:
                num_eus_for_core = core_block['eu_num']
                break
        if not num_eus_for_core:
            raise HandlerException(f"Core id {core} for {eug_name} of partition {key} "
                                   "does not exist in DPA resources")
        if eus > num_eus_for_core:
            raise HandlerException(f"EU number {eus} for {eug_name} of "
                                   f"partition {key} exceeds "
                                   f"number of EU {num_eus_for_core}")
        start_subeu = 0
        subeu = 0
        while subeu < eus:
            if start_subeu + subeu >= num_eus_for_core:
                raise HandlerException("Number of eus more than "
                                       f"{num_eus_for_core} in "
                                       f"{eug_name} of partition {key}")
            eu = core_block['eus'][start_subeu + subeu]['eu_id']
            if eu not in self._eu_map:
                if key != ROOT_PARTITION_NAME:
                    start_subeu += 1
                    continue
                self._eu_map[eu] = self._dict_for_eu(key, eu)

            if self._eu_map[eu]['partition_name'] != key:
                start_subeu += 1
                continue
            if self._eu_map[eu]['eug'] != "":
                start_subeu += 1
                continue
            self._eu_map[eu]['eug'] = eug_name
            eug_key = (key, eug_name)
            if eug_key not in self._eug_db:
                self._eug_db[eug_key] = {}
            subeu += 1

    # Function for map groups of partition
    def _fill_map_groups(self, partition_name, key_name, eug_allocation_list):
        if not eug_allocation_list:
            return

        eug_names = []

        for eug_allocation in eug_allocation_list:
            try:
                eug_name = eug_allocation[NAME]
            except KeyError as exc:
                raise HandlerException(f"Absent EUG's {NAME} in {key_name} of "
                                       f"partition {partition_name}") from exc
            if not eug_name:
                raise HandlerException(f"Empty EUG's {NAME} in {key_name} of "
                                       f"partition {partition_name}")

            if len(eug_name) > EUG_MAXIMAL_LENGTH:
                raise HandlerException(f"EUG {NAME} in {key_name} of partition {partition_name} "
                                       "exceed the maximum allowed length of "
                                       f"{EUG_MAXIMAL_LENGTH} characters")

            if eug_name in eug_names:
                raise HandlerException(f"Duplicate EUG name {eug_name} in {key_name} of "
                                       f"partition {partition_name}")
            eug_names.append(eug_name)

            try:
                temp_allocation = eug_allocation[EUS]
            except KeyError as exc:
                raise HandlerException(f"Absent EUG's {EUS} field in {key_name} "
                                       f"of partition {partition_name}") from exc

            if not isinstance(temp_allocation, (dict, list)):
                raise HandlerException(f"Invalid type for EUG's {eug_name} {EUS} in {key_name} "
                                       f"of partition {partition_name}. Expected dict or list.")

            allocation = convert_values_to_list(temp_allocation, EUS, partition_name)

            for elem in allocation:
                self._fill_eu_map_for_eug(elem, partition_name, eug_name)

    # Function to fill EU map
    def _fill_eu_map(self):

        if PARTITIONS in self._main_db:
            self._partition_db = self._main_db[PARTITIONS]

        for key in self._partition_db:
            allocation = get_value_as_list(self._partition_db, key,
                                           EU_ALLOCATION)
            for elem in allocation:
                self._fill_eu_map_for_eu(elem, key)

            alloc_str = EU_GROUP_ALLOCATION
            eug_allocation_list = get_value_as_list(self._partition_db, key,
                                                    alloc_str, False)

            self._fill_map_groups(key, alloc_str, eug_allocation_list)

        if ROOT_PARTITION_GROUPS not in self._main_db:
            return

        eug_allocation_list = self._main_db[ROOT_PARTITION_GROUPS]
        self._fill_map_groups(ROOT_PARTITION_NAME, ROOT_PARTITION_GROUPS, eug_allocation_list)

    # Function to fill database of EUs for applications
    def _init_map_for_apps(self, eu):
        try:
            self._eu_map[eu]['apps'] = []
        except KeyError as exc:
            raise HandlerException(f"EU {eu} is not allocated for any partition or group {exc}") \
                from exc

    # Check database of EU for apps with core/num_EUs pair values
    def _check_core_eu_for_apps(self, eu_block, partition_name, app_name, sub_dict):

        if isinstance(eu_block, dict):
            eu_list = [eu_block]
        else:
            eu_list = eu_block

        dict_eu = {}

        for elem in eu_list:

            core, eus = self._check_core_eus(elem, partition_name, app_name=app_name)

            num_eus_for_core = None
            core_block = None
            for core_block in self._res_data.cores:
                if core_block['core_id'] == core:
                    num_eus_for_core = core_block['eu_num']
                    break
            if not num_eus_for_core:
                raise HandlerException(f"Core id {core} for partition {partition_name} in "
                                       f"application {app_name} does not exist in the "
                                       "available DPA resources")
            if eus > num_eus_for_core:
                raise HandlerException(f"The number of EUS {eus} for partition {partition_name} of "
                                       f"application {app_name} exceeds the total number of EUs "
                                       f"{num_eus_for_core} available for core {core}")

            start_subeu = 0
            subeu = 0
            while subeu < eus:
                if start_subeu + subeu >= num_eus_for_core:
                    raise HandlerException(f"The number of EUs specified as {eus} for core {core} "
                                           f"in partition {partition_name} of "
                                           f"application {app_name} exceeds the number of "
                                           "available EUs for this core within the partition.")
                eu = core_block['eus'][start_subeu + subeu]['eu_id']
                if (eu in sub_dict[AFFINITY_EUS]) or (eu in dict_eu):
                    start_subeu += 1
                    continue
                if eu not in self._eu_map:
                    if partition_name != ROOT_PARTITION_NAME:
                        start_subeu += 1
                        continue
                    self._eu_map[eu] = self._dict_for_eu(partition_name, eu)

                if self._eu_map[eu]['partition_name'] != partition_name:
                    start_subeu += 1
                    continue
                subeu += 1
                dict_eu[eu] = self._eu_map[eu]['rel_idx']
                self._init_map_for_apps(eu)

        if len(dict_eu) > 0:
            sub_dict[AFFINITY_EUS].extend(list(dict_eu.keys()))
            sub_dict[RELATIVE_EUS].extend(list(dict_eu.values()))

    # Fill database of EU for apps with core/num_EUs pair values
    def _fill_core_eu_for_apps(self, eu_block, partition_name, app_name, sub_dict):

        if isinstance(eu_block, dict):
            eu_list = [eu_block]
        else:
            eu_list = eu_block

        dict_eu = {}

        for elem in eu_list:

            core, eus = self._check_core_eus(elem, partition_name, app_name=app_name)

            sub_dict_eu = {}
            num_eus_for_core = None
            core_block = None
            for core_block in self._res_data.cores:
                if core_block['core_id'] == core:
                    num_eus_for_core = core_block['eu_num']
                    break

            for eu_index in range(num_eus_for_core):
                eu = core_block['eus'][eu_index]['eu_id']
                # EU is unassigned
                if eu not in self._eu_map:
                    continue
                # EU in another partition
                if partition_name != self._eu_map[eu]['partition_name']:
                    continue
                try:
                    if app_name in self._eu_map[eu]['apps']:
                        continue
                except KeyError:
                    self._init_map_for_apps(eu)
                sub_dict_eu[eu] = len(self._eu_map[eu]['apps'])

            list_eus = sorted(sub_dict_eu, key=sub_dict_eu.get)
            list_eus = list_eus[:eus]
            for eu in list_eus:
                dict_eu[eu] = self._eu_map[eu]['rel_idx']
                self._eu_map[eu]['apps'].append(app_name)

        if len(dict_eu) > 0:
            sub_dict[AFFINITY_EUS].extend(list(dict_eu.keys()))
            sub_dict[RELATIVE_EUS].extend(list(dict_eu.values()))

    # Fill database of EU for apps with absolute EU values
    def _fill_abs_eu_for_apps(self, eu_block, partition_name, app_name, check, sub_dict):
        dict_eu = {}
        eu_list_elem = convert_values_to_list(eu_block, EU_ALLOCATION, app_name)
        list_eus = []
        for eu_sub_block in eu_list_elem:
            str_eu_sub_block = str(eu_sub_block)
            list_eus.append(str_eu_sub_block.replace(" ", ""))
        try:
            eu_list = expand_range_from_string(",".join(list_eus), suppress_value_error_exc=False)
        except ValueError as exc:
            raise HandlerException(f"Error expanding EU range '{','.join(list_eus)}' "
                                   f"for partition {partition_name} in {app_name}") from exc
        for eu in eu_list:
            if (eu in sub_dict[AFFINITY_EUS]) or (eu in dict_eu):
                raise HandlerException(f"EU {eu} for partition {partition_name} in application "
                                       f"{app_name} is already taken by this application")
            try:
                eu_partition_name = self._eu_map[eu]['partition_name']
            except KeyError as exc:
                if partition_name != ROOT_PARTITION_NAME:
                    raise HandlerException(f"Absent EU {eu} for partition "
                                           f"{partition_name} in {app_name}") from exc
                self._eu_map[eu] = self._dict_for_eu(partition_name, eu)
                eu_partition_name = partition_name

            if eu_partition_name != partition_name:
                raise HandlerException(f"EU {eu} for partition "
                                       f"{partition_name} in {app_name} "
                                       f"mapped on {eu_partition_name}")

            if check:
                self._init_map_for_apps(eu)
            dict_eu[eu] = self._eu_map[eu]['rel_idx']
            if not check:
                self._eu_map[eu]['apps'].append(app_name)

        if len(dict_eu) > 0:
            sub_dict[AFFINITY_EUS].extend(list(dict_eu.keys()))
            sub_dict[RELATIVE_EUS].extend(list(dict_eu.values()))

    # Fill database of groups for applications
    def _fill_eug_for_apps(self, eug_block, curr_partition, app_name, sub_dict):
        dict_check = {}
        eug_list = convert_values_to_list(eug_block, EUS, app_name)
        list_eugs = []
        for eug_sub_block in eug_list:
            list_eugs.extend(eug_sub_block.replace(" ", "").split(","))
        for eug in list_eugs:
            eug_key = (curr_partition, eug)
            if eug_key in dict_check:
                raise HandlerException(f"The group {eug} already exist for "
                                       f"partition {curr_partition} "
                                       f"in {app_name}")
            if eug_key not in self._eug_db:
                raise HandlerException(f"Unknown group {eug} for partition "
                                       f"{curr_partition} in {app_name}")
            dict_check[eug_key] = self._eug_db[eug_key]

        if len(dict_check) == 0:
            del sub_dict[EU_GROUP_ALLOCATION]
        else:
            sub_dict[EU_GROUP_ALLOCATION] = dict_check

    # Fill apps database
    def _fill_apps(self, check=False):
        if DPA_APPS not in self._main_db:
            return
        dpa_apps = self._main_db[DPA_APPS]
        if not isinstance(dpa_apps, (dict)):
            raise HandlerException(f"The '{DPA_APPS}' field has an invalid type. "
                                   "A mapping (dictionary) is expected.")
        for app_name, app in dpa_apps.items():
            list_app = convert_values_to_list(app, app_name, DPA_APPS)
            partition_list = []
            for elem_app in list_app:
                if VERSION in elem_app:
                    raise HandlerException(f"Forbidden field '{VERSION}' "
                                           f"for application {app_name}")
                if PARTITION not in elem_app:
                    raise HandlerException(f"Absent '{PARTITION} "
                                           f"for application {app_name}")
                partition_name = elem_app[PARTITION]
                if partition_name not in self._main_db[PARTITIONS] and partition_name != 'ROOT':
                    raise HandlerException(f"Unknown '{PARTITION}' name "
                                           f"{partition_name} for application "
                                           f"{app_name}")

                if partition_name in partition_list:
                    raise HandlerException(f"'{PARTITION}' name {partition_name} for application "
                                           f"{app_name} is already used.")

                partition_list.append(partition_name)

                if partition_name not in self._apps_db:
                    self._apps_db[partition_name] = []

                sub_dict = {NAME: app_name, EU_GROUP_ALLOCATION: [],
                            AFFINITY_EUS: [], RELATIVE_EUS: []}

                if EU_GROUP_ALLOCATION in elem_app:
                    if not isinstance(elem_app[EU_GROUP_ALLOCATION], (int, str, list)):
                        raise HandlerException(f"Invalid type for {EU_GROUP_ALLOCATION} in "
                                               f"application {app_name} for {PARTITION} "
                                               f"{partition_name}.")
                    self._fill_eug_for_apps(elem_app[EU_GROUP_ALLOCATION],
                                            partition_name, app_name, sub_dict)
                if AFFINITY_EUS in elem_app:
                    if not isinstance(elem_app[AFFINITY_EUS], (int, str, list)):
                        raise HandlerException(f"Invalid type for {AFFINITY_EUS} in "
                                               f"application {app_name} for {PARTITION} "
                                               f"{partition_name}. Expected int, str, or list.")
                    self._fill_abs_eu_for_apps(elem_app[AFFINITY_EUS], partition_name,
                                               app_name, check, sub_dict)
                if AFFINITY_CORE in elem_app:
                    if not isinstance(elem_app[AFFINITY_CORE], (dict, list)):
                        raise HandlerException(f"Invalid type for {AFFINITY_CORE} in "
                                               f"application {app_name} for {PARTITION} "
                                               f"{partition_name}.")
                    if check:
                        self._check_core_eu_for_apps(elem_app[AFFINITY_CORE],
                                                     partition_name, app_name, sub_dict)
                    else:
                        self._fill_core_eu_for_apps(elem_app[AFFINITY_CORE],
                                                    partition_name, app_name, sub_dict)
                if (len(sub_dict[EU_GROUP_ALLOCATION]) == 0) and (len(sub_dict[AFFINITY_EUS]) == 0):
                    raise HandlerException(f"Application {app_name} in partition {partition_name} "
                                           "has no EUs or groups.")
                if not check:
                    self._apps_db[partition_name].append(sub_dict)

    # Function for return correct string of partitions EUs and groups
    def _create_string_eu(self, key, eug=None):
        eu_list = []
        eug_list = []
        rel_idx = 0
        first = last = curr = INDEX_INVALID_VALUE
        for elem in self._eu_map.values():
            if key != elem['partition_name']:
                if first != INDEX_INVALID_VALUE:
                    eu_list.append(create_substring_in_range(first, last))
                    first = INDEX_INVALID_VALUE
                continue
            if eug and eug != elem['eug']:
                continue
            if elem['eug'] not in eug_list and elem['eug'] != '':
                eug_list.append(elem['eug'])
            curr = elem['eu']
            if first == INDEX_INVALID_VALUE:
                first = curr
            elif last + 1 < curr:
                if first != INDEX_INVALID_VALUE:
                    eu_list.append(create_substring_in_range(first, last))
                first = curr
            last = curr
            elem['rel_idx'] = rel_idx
            rel_idx += 1
        if first != INDEX_INVALID_VALUE:
            eu_list.append(create_substring_in_range(first, last))
            first = INDEX_INVALID_VALUE

        return ",".join(eu_list), eug_list

    # Function for check validity of vhca_id and add them to
    # partitions as string
    def _check_and_add_vhca_ids(self):
        for key, partition in self._main_db[PARTITIONS].items():
            vhca_ids = partition[VHCA_ID]
            raw_vhca_id_list = convert_values_to_list(vhca_ids, {key}, VHCA_ID)
            try:
                vhca_int_list = expand_range_from_string(",".join(raw_vhca_id_list),
                                                         suppress_value_error_exc=False)
            except ValueError as exc:
                raise HandlerException(f"Error processing {VHCA_ID} "
                                       f"for partition {key}: {exc}") from exc
            for vhca_id in vhca_int_list:
                if vhca_id not in self._vhca_data.obj_dict:
                    raise HandlerException(f"vhca_id {vhca_id} in partition "
                                           f"{key} does not exist")
                function_type = self._vhca_data.obj_dict[vhca_id]['function_type']
                if function_type == 3:
                    raise HandlerException(f"vhca_id {vhca_id} in partition "
                                           f"{key} is a root and cannot be used")
                if function_type == 2:
                    raise HandlerException(f"vhca_id {vhca_id} in partition "
                                           f"{key} is a SF and cannot be used")
                if 'used' in self._vhca_data.obj_dict[vhca_id]:
                    raise HandlerException(f"vhca_id {vhca_id} in partition "
                                           f"{key} is already used in partition "
                                           f"{self._vhca_data.obj_dict[vhca_id]['used']}")
                self._vhca_data.obj_dict[vhca_id]['used'] = key
            len_list = len(vhca_int_list)
            max_len = self._res_data.max_vhca_id_size
            if len_list > max_len:
                raise HandlerException(f"length of vhca_id {len(len_list)} "
                                       "exceeded maximal number of vhca_id "
                                       f"vhca_id  for partition {max_len}")
            vhca_string = create_string_from_list(vhca_int_list)
            partition['vhca_id_string'] = vhca_string

    # Function for run partition create for each partition
    def _run_commands_for_partition(self):
        for key, partition in self._main_db[PARTITIONS].items():
            eu_ids, eu_group = self._create_string_eu(key)
            out_str = self._chip_data.run_command("partition", "create",
                                                  vhca_list=partition['vhca_id_string'],
                                                  range_eus=eu_ids,
                                                  max_num_eu_group=len(eu_group))
            value = get_value_from_list_of_strings(out_str, ["Partition", "ID"])
            if not value:
                raise HandlerException(f"Invalid output for partition {key}")
            self._main_db[PARTITIONS][key]['ID'] = value

    # Function for run group create
    def _run_commands_for_groups(self):
        # pylint: disable=consider-using-dict-items
        for key in self._eug_db:
            (partition_name, group_name) = key
            eu_list = [val['rel_idx'] for val in self._eu_map.values()
                       if (val['partition_name'] == partition_name and
                           val['eug'] == group_name)]
            range_eu = create_string_from_list(eu_list)
            if partition_name != ROOT_PARTITION_NAME:
                parttion_id = self._partition_db[partition_name]['ID']
                out_str = self._chip_data.run_command("eu_group", "create",
                                                      range=range_eu,
                                                      id_partition=parttion_id,
                                                      name_group=group_name)
            else:
                out_str = self._chip_data.run_command("eu_group", "create",
                                                      range=range_eu,
                                                      name_group=group_name)
            group_id = get_value_from_list_of_strings(out_str, ["group", "ID"])
            if not group_id:
                raise HandlerException(f"Invalid output for group {group_name} in "
                                       f"partition {partition_name}")
            self._eug_db[key]['ID'] = group_id
            self._eug_db[key]['range'] = range_eu
        # pylint: enable=consider-using-dict-items

    # Function for collecting information for yaml files
    def _collect_information_for_yamls(self):
        output_files = list(self._main_db[PARTITIONS].keys())
        output_files.append(ROOT_PARTITION_NAME)
        for file_name in output_files:
            yaml_db = {}
            try:
                yaml_db[DPA_APPS] = self._apps_db[file_name]
            except KeyError:
                yaml_db[DPA_APPS] = []
            if len(yaml_db[DPA_APPS]):
                self._yaml_obj.write_yaml_file(file_name, yaml_db, self._options.output)

    # Read information from the chip
    def _read_info_from_chip(self):
        self._chip_data.read_vhca_data()
        self._chip_data.read_dpa_resources()
        self._chip_data.read_dpa_partitions_resources()
        self._chip_data.read_dpa_groups_resources()
        self._vhca_data = VHCATreeData(self._chip_data.vhca_data)
        self._res_data = DPAResources(self._chip_data.resource_data)

    # Read status of the chip
    def _check_status(self):
        out_dict = self._chip_data.get_status()
        for key, value in out_dict.items():
            if value:
                raise HandlerException(f"Error in status check: the current {key} is {value} "
                                       f"but should be 0.\n"
                                       "Please destroy all active DPA elements or reset the chip, "
                                       "then restart the application.")

    # Function for basic checking of yaml file
    def _base_checking(self):
        partitions = self._main_db[PARTITIONS]
        if not isinstance(partitions, (dict)):
            raise HandlerException(f"The '{PARTITIONS}' field has an invalid type. "
                                   "A mapping (dictionary) is expected.")
        partitions_num = len(partitions)
        max_num_partitions = self._res_data.max_num_partitions
        if len(partitions) > max_num_partitions:
            raise HandlerException(f"Number of partitions {partitions_num} "
                                   f"exceeds max_num_partitions {max_num_partitions}")
        for partition_name, partition in partitions.items():
            if not isinstance(partition, (dict)):
                raise HandlerException(f"The partition {partition_name} field has "
                                       "an invalid type. A mapping (dictionary) is expected.")
            if VHCA_ID not in partition:
                raise HandlerException(f"Absent key '{VHCA_ID}' for "
                                       f"partition {partition_name}")
            if not isinstance(partition[VHCA_ID], (int, list, str)):
                raise HandlerException(f"The '{VHCA_ID}' field has an invalid type for "
                                       f"partition {partition_name}. "
                                       "It must be an int, list, or str.")
            if EU_ALLOCATION not in partition:
                raise HandlerException(f"Absent key '{EU_ALLOCATION}' for "
                                       f"partition {partition_name}")
            if not isinstance(partition[EU_ALLOCATION], (dict, list)):
                raise HandlerException(f"The '{EU_ALLOCATION}' field has an invalid type for "
                                       f"partition {partition_name}. "
                                       "A mapping (dictionary or list) is expected.")
        group_num = 0
        for partition_name, elem in partitions.items():
            if EU_GROUP_ALLOCATION in elem:
                group = elem[EU_GROUP_ALLOCATION]
                if not isinstance(group, (dict, list)):
                    raise HandlerException(f"The '{EU_GROUP_ALLOCATION}' field has an invalid "
                                           f"type for partition {partition_name}. "
                                           "A mapping (dictionary or list) is expected.")
                group_num += len(group)

        if ROOT_PARTITION_GROUPS in self._main_db:
            group = self._main_db[ROOT_PARTITION_GROUPS]
            if not isinstance(group, (dict, list)):
                raise HandlerException(f"The '{ROOT_PARTITION_GROUPS}' field has an invalid type. "
                                       f"A mapping (dictionary or list) is expected.")
            group_num += len(group)

        max_groups = self._res_data.max_eu_groups
        if group_num + partitions_num + 1 > max_groups:
            raise HandlerException(f"Number of group {group_num} + number of "
                                   f"partitions {partitions_num} + 1 reserved exceeds "
                                   f"max_eu_groups {max_groups}")

    # Main function for work
    def run(self):
        self._yaml_obj.process_data()
        self._chip_data.get_tool_version()
        self._check_status()
        self._read_info_from_chip()
        self._main_db = self._yaml_obj.get_yaml_database()
        self._base_checking()
        self._fill_eu_map()
        self._check_and_add_vhca_ids()
        self._fill_apps(True)
        self._run_commands_for_partition()
        self._run_commands_for_groups()
        self._fill_apps()
        self._collect_information_for_yamls()
