#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Base module for check input database, input from the chip, create partitions
# and group on the chip, and create output yaml files.


import re
import sys
import io
from ..databases.vhcatree import VHCATreeData
from ..databases.dparesource import DPAResources
from ..utilities.string import expand_range_from_string, create_string_from_list
from ..constants.constants import PARTITIONS, ROOT_PARTITION_GROUPS, VERSION, NAME, \
    EU_ALLOCATION, ABS_EUS, EU_GROUP_ALLOCATION, EUS, DEFAULT_OUTPUT_YAML_VERSION


# Function for get string with information about core
def _get_string_with_core_info(core):
    coreid = f"Core #{core['core_id']}, with {core['eu_num']} EUs, "
    if core['empty'] == 0:
        core_data = "is fully utilized."
    elif core['empty'] == core['eu_num']:
        core_data = "is fully available."
    elif core['empty'] == 1:
        core_data = f"has one EU available: {core['eu_list'][0]}."
    else:
        string_lst = create_string_from_list(core['eu_list'])
        core_data = f"has {core['empty']} EUs available: {string_lst}."
    return f"{coreid}{core_data}"


# Main class for query


class QueryOnlyHandler:
    def __init__(self, **kwargs):
        self._options = kwargs['parse_options']
        self._chip_data = kwargs['exchange_obj'](self._options.device, debug=self._options.verbose)
        self._elements = {VERSION: DEFAULT_OUTPUT_YAML_VERSION,
                          PARTITIONS: {},
                          ROOT_PARTITION_GROUPS: []}
        self._internal_data = {}
        self._offset = 4
        self._errors_in_core_mapping = True
        self._free_vhca = []

    def open_file_for_write(self):
        if not self._options.filename:  # Use stdout by default
            sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')
            output = sys.stdout
        else:
            try:
                # pylint: disable=consider-using-with
                output = open(self._options.filename, 'w', encoding="utf-8")
                # pylint: enable=consider-using-with
            # pylint: disable=bare-except
            # pylint: disable=broad-except
            except BaseException:
                print(f"Cannot open {self._options.filename} for write")
                sys.exit(1)
            # pylint: enable=broad-except
            # pylint: enable=bare-except
        return output

    # Function to parse partition query
    def _parse_partition_query(self, text):

        out_string = "\n".join(text)
        pattern = r"(\d+)\) EU Partition ID: (\d+)\n" \
                  r"Maximal number of groups: (\d+)\n" \
                  r"The partition has a total of \d+ associated VHCA IDs, namely: ([\d,-]+)\n" \
                  r"Partition's member EUs are: ([\d,-]+)"

        partition_pattern = re.findall(pattern, out_string)

        for partition in partition_pattern:
            partition_dict = {}
            partition_dict['partition_nmb'] = int(partition[0])
            partition_dict['partition_id'] = int(partition[1])
            partition_dict['max_groups'] = int(partition[2])
            partition_name = "PARTITION_" + str(partition_dict['partition_nmb'])
            partition_dict['vhca_ids'] = expand_range_from_string(partition[3])
            partition_dict['range_eus'] = expand_range_from_string(partition[4])
            for eu in partition_dict['range_eus']:
                self._internal_data['eu_map'][eu]['partition'] = partition_name
            if partition_dict['max_groups']:
                out_list = self._chip_data.read_dpa_groups_query(partition_dict['partition_id'])
                partition_dict['groups'] = self._parse_group_query(out_list,
                                                                   partition_dict['range_eus'])
            self._elements['partitions'][partition_name] = partition_dict

    # Function to parse eu group query
    def _parse_group_query(self, text, range_eus=None, root_group=False):

        out_string = "\n".join(text)
        if root_group:
            pattern = r"(\d+)\) EU group ID: (\d+)\n" \
                      r"EU group name: (\S+)\n" \
                      r"Member EUs are: ([\d,-]+)\n"
        else:
            pattern = r"(\d+)\) EU group ID: (\d+)\n" \
                      r"EU group name: (\S+)\n" \
                      r"Partition ID: \S+\n" \
                      r"Member EUs are: ([\d,-]+)\n"

        group_list = []

        group_pattern = re.findall(pattern, out_string)

        for group in group_pattern:
            group_dict = {}
            group_dict['group_nmb'] = int(group[0])
            group_dict['group_id'] = int(group[1])
            group_dict[NAME] = group[2]
            group_dict['range_eus'] = expand_range_from_string(group[3])
            if range_eus:
                group_dict['abs_range_eus'] = []
                abs_range_eus = {}
                for index, eu in enumerate(range_eus):
                    abs_range_eus[index] = eu
                    self._internal_data['eu_map'][eu]['eu_group'] = group_dict[NAME]
                for eu in group_dict['range_eus']:
                    group_dict['abs_range_eus'].append(abs_range_eus[eu])
            else:
                for index, eu in enumerate(group_dict['range_eus']):
                    self._internal_data['eu_map'][eu]['eu_group'] = group_dict[NAME]
            group_list.append(group_dict)

        return group_list

    # Function to read and parse status
    def _read_and_parse_status(self):
        self._chip_data.read_dpa_groups_resources()
        self._chip_data.read_dpa_partitions_resources()
        self._chip_data.read_dpa_resources()
        self._internal_data['resource_data'] = DPAResources(self._chip_data.resource_data)
        self._chip_data.read_vhca_data()
        self._internal_data['vhca_data'] = VHCATreeData(self._chip_data.vhca_data)
        max_num_dpa_eu = self._internal_data['resource_data'].max_num_dpa_eu
        invalid_core_id = 13000
        self._internal_data['eu_map'] = [{'partition': None, 'eu_group': None,
                                          'core_id': invalid_core_id}
                                         for _ in range(max_num_dpa_eu)]
        cores = self._internal_data['resource_data'].cores
        errors_great = False
        errors_reassing = False
        for core in cores:
            for eu_dict in core['eus']:
                eu_id = eu_dict['eu_id']
                core_id = eu_dict['core_id']
                if eu_id >= max_num_dpa_eu:
                    if not errors_great:
                        print(f"eu_id {eu_id} in core {core_id} greater than {max_num_dpa_eu}")
                    errors_great = True
                    continue
                if self._internal_data['eu_map'][eu_id]['core_id'] != invalid_core_id:
                    if not errors_reassing:
                        print(f"eu_id {eu_id} in core {core_id} already assigned to "
                              f"{self._internal_data['eu_map'][eu_id]['core_id']}")
                    errors_reassing = True
                    continue
                self._internal_data['eu_map'][eu_id]['core_id'] = core_id

        error_not_assigned = False
        for index, eu in enumerate(self._internal_data['eu_map']):
            if eu['core_id'] == invalid_core_id:
                print(f"EU {index} not assigned by any core")
                error_not_assigned = True
                break
        if not (errors_great or errors_reassing or error_not_assigned):
            self._errors_in_core_mapping = False

        out_list = self._chip_data.read_dpa_partitions_query()
        self._parse_partition_query(out_list)
        out_list = self._chip_data.read_dpa_groups_query()
        self._elements[ROOT_PARTITION_GROUPS] = self._parse_group_query(out_list, root_group=True)
        self._elements['pure_eus'] = [index for index, eu in
                                      enumerate(self._internal_data['eu_map'])
                                      if eu['partition'] is None and eu['eu_group'] is None]
        self._elements['free_eus'] = [index for index, eu in
                                      enumerate(self._internal_data['eu_map'])
                                      if eu['partition'] is None]

    # Function to create core ids list
    def _create_core_ids_list(self, eus_list):
        core_ids = []
        if len(eus_list) == 0:
            return core_ids
        for core in self._internal_data['resource_data'].cores:
            for eu in core['eus']:
                if eu['eu_id'] in eus_list and eu['core_id'] not in core_ids:
                    core_ids.append(eu['core_id'])
        return core_ids

    # Function to create eu list
    def _create_eu_list(self, check_group=True, partition=None, group=None):
        if check_group:
            return [index for index, eu in enumerate(self._internal_data['eu_map'])
                    if eu['partition'] == partition and eu['eu_group'] == group]
        return [index for index, eu in enumerate(self._internal_data['eu_map'])
                if eu['partition'] == partition]

    # Function to create eu list
    def _calculate_free_vhca(self):
        work_vhca = [vhca_id for vhca_id, vhca_elem in
                     self._internal_data['vhca_data'].obj_dict.items()
                     if vhca_elem['function_type'] == 0]
        for val in self._elements['partitions'].values():
            for vhca_id in val['vhca_ids']:
                if vhca_id in work_vhca:
                    work_vhca.remove(vhca_id)
        self._free_vhca = work_vhca

    # Function to print partition root header
    def _print_partition_root_header(self, output):
        eus_list0 = self._create_eu_list()
        eus_list1 = None
        if len(eus_list0):
            eus_list1 = create_string_from_list(eus_list0)
            output.write(f"# Free EUs in ROOT partition {eus_list1}\n")
            core_id_list0 = self._create_core_ids_list(eus_list0)
            if len(core_id_list0) > 0:
                core_id_list1 = create_string_from_list(core_id_list0)
                output.write(f"# Cores available for use in ROOT partition {core_id_list1}\n")
        eus_list0 = self._create_eu_list(False)
        eus_list2 = create_string_from_list(eus_list0)
        if eus_list1 and (len(eus_list2) > 0) and (eus_list1 != eus_list2):
            output.write(f"# Free EUs in ROOT partition with groups {eus_list2}\n")
            core_id_list0 = self._create_core_ids_list(eus_list0)
            if len(core_id_list0) > 0:
                core_id_list2 = create_string_from_list(core_id_list0)
                output.write("# Cores available for use ROOT partition with "
                             f"groups {core_id_list2}\n")

    # Function to print partition root
    def _print_partition_root(self, output):
        eus_list0 = self._create_eu_list()
        eus_list1 = None
        if len(eus_list0):
            eus_list1 = create_string_from_list(eus_list0)
            output.write(f"Free EUs in ROOT partition {eus_list1}\n")
            core_id_list0 = self._create_core_ids_list(eus_list0)
            if len(core_id_list0) > 0:
                core_id_list1 = create_string_from_list(core_id_list0)
                output.write(f"Cores available for use in ROOT partition {core_id_list1}\n")
        eus_list0 = self._create_eu_list(False)
        eus_list2 = create_string_from_list(eus_list0)
        if eus_list1 and (len(eus_list2) > 0) and (eus_list1 != eus_list2):
            output.write(f"Free EUs in ROOT partition with groups {eus_list2}\n")
            core_id_list0 = self._create_core_ids_list(eus_list0)
            if len(core_id_list0) > 0:
                core_id_list2 = create_string_from_list(core_id_list0)
                output.write("Cores available for use ROOT partition with "
                             f"groups {core_id_list2}\n")

    # Function to print partition as YAML
    def _print_partition_as_yaml(self, output, partition):
        output.write(f"    # partition_id: {partition['partition_id']}\n")
        output.write(f"    vhca_id: {create_string_from_list(partition['vhca_ids'])}\n")
        output.write(f"    {EU_ALLOCATION}:\n")
        range_eus = create_string_from_list(partition['range_eus'])
        output.write(f"      - {ABS_EUS}: {range_eus}\n")
        core_id_list0 = self._create_core_ids_list(partition['range_eus'])
        if len(core_id_list0) > 0:
            core_id_list1 = create_string_from_list(core_id_list0)
            output.write(f"      # Cores in use: {core_id_list1}\n")

    # Function to print group as YAML
    def _print_group_as_yaml(self, output, group, prefix):
        abs_range = 'abs_range_eus' in group
        output.write(f"{prefix}- {NAME}: {group[NAME]}\n")
        output.write(f"{prefix} # group_id: {group['group_id']}\n")
        output.write(f"{prefix}  {EUS}:\n")
        key_range = 'abs_range_eus' if abs_range else 'range_eus'
        range_eus = create_string_from_list(group[key_range])
        output.write(f"{prefix}    - {ABS_EUS}: {range_eus}\n")
        if abs_range:
            range_eus = create_string_from_list(group['range_eus'])
            output.write(f"{prefix}    # relative_EUs: {range_eus}\n")
        core_id_list0 = self._create_core_ids_list(group[key_range])
        if len(core_id_list0) > 0:
            core_id_list1 = create_string_from_list(core_id_list0)
            output.write(f"{prefix}    # Cores in use: {core_id_list1}\n")

    # Function to print status as YAML
    def _print_status_as_yaml(self, output):
        output.write("---\n")
        output.write(f"{VERSION}: {self._elements[VERSION]}\n")
        self._print_partition_root_header(output)
        self._print_yaml_core_mapping(output)
        self._print_yaml_vhca_mapping(output)
        if len(self._elements[PARTITIONS]) > 0:
            output.write(f"{PARTITIONS}:\n")
            for name, partition in self._elements[PARTITIONS].items():
                output.write(f"  {name}:\n")
                self._print_partition_as_yaml(output, partition)
                groups_nmb = len(partition['groups']) if 'groups' in partition else 0
                if partition['max_groups'] > groups_nmb:
                    output.write(f"    # max_groups: {partition['max_groups']}\n")
                if groups_nmb > 0:
                    output.write(f"    {EU_GROUP_ALLOCATION}:\n")
                    for group in partition['groups']:
                        self._print_group_as_yaml(output, group, " " * 6)
        else:
            output.write(f"{PARTITIONS}: {{}}\n")
        if len(self._elements[ROOT_PARTITION_GROUPS]) > 0:
            output.write(f"{ROOT_PARTITION_GROUPS}:\n")
            for group in self._elements[ROOT_PARTITION_GROUPS]:
                self._print_group_as_yaml(output, group, " " * 2)
        output.write("---\n")

    def _calculate_cores(self):
        core_list = []
        eu_map = self._internal_data['eu_map']
        cores = self._internal_data['resource_data'].cores
        for core in cores:
            core_inf = {}
            eu_list = [val['eu_id'] for val in core['eus']
                       if eu_map[val['eu_id']]['partition'] is None and
                       eu_map[val['eu_id']]['eu_group'] is None]
            core_inf['empty'] = len(eu_list)
            core_inf['core_id'] = core['core_id']
            core_inf['eu_num'] = core['eu_num']
            core_inf['eu_list'] = eu_list
            core_list.append(core_inf)
        return core_list

    def _print_core_mapping(self, output):
        core_inf = self._calculate_cores()
        output.write("\n")
        output.write("Core's information:\n")
        len_core_inf = len(core_inf)
        for index, core in enumerate(core_inf):
            prefix = self._get_connector(index >= (len_core_inf - 1))
            core_data = _get_string_with_core_info(core)
            output.write(f"{prefix}{core_data}\n")

    def _print_yaml_core_mapping(self, output):
        core_inf = self._calculate_cores()
        output.write("# Core's information:\n")
        for core in core_inf:
            core_data = _get_string_with_core_info(core)
            output.write(f"# {core_data}\n")

    def _print_vhca_mapping(self, output):
        self._calculate_free_vhca()
        len_free = len(self._free_vhca)
        output.write("\n")
        output.write("VHCA information:\n")
        if len_free > 0:
            output.write(f"    {len_free} unutilized VHCAs: "
                         f"{create_string_from_list(self._free_vhca)}\n")
        else:
            output.write("    All VHCA resources are fully utilized.\n")

    def _print_yaml_vhca_mapping(self, output):
        self._calculate_free_vhca()
        len_free = len(self._free_vhca)
        if len_free > 0:
            output.write(f"# {len_free} unutilized VHCAs: "
                         f"{create_string_from_list(self._free_vhca)}\n")
        else:
            output.write("# All VHCA resources are fully utilized.\n")

    # Function to print partition
    def _print_partition(self, output, partition, prefix):
        output.write(f"{prefix}partition_id: {partition['partition_id']}\n")
        output.write(f"{prefix}vhca_id: {create_string_from_list(partition['vhca_ids'])}\n")
        range_eus = create_string_from_list(partition['range_eus'])
        output.write(f"{prefix}Range of EUs: {range_eus}\n")
        core_id_list0 = self._create_core_ids_list(partition['range_eus'])
        if len(core_id_list0) > 0:
            core_id_list1 = create_string_from_list(core_id_list0)
            output.write(f"{prefix}Cores in use: {core_id_list1}\n")

    # Function to print group
    def _print_group(self, output, group, base_prefix, is_last):
        abs_range = 'abs_range_eus' in group

        prefix = f"{base_prefix}{self._get_space(False)}"
        output.write(f"{prefix}\n")
        connector = self._get_connector(is_last)
        space = self._get_space(is_last)

        output.write(f"{base_prefix}{connector}{NAME}: {group[NAME]}\n")

        prefix = f"{base_prefix}{space}"

        output.write(f"{prefix}group_id: {group['group_id']}\n")
        key_range = 'abs_range_eus' if abs_range else 'range_eus'
        range_eus = create_string_from_list(group[key_range])
        output.write(f"{prefix}{ABS_EUS}: {range_eus}\n")
        if abs_range:
            range_eus = create_string_from_list(group['range_eus'])
            output.write(f"{prefix}Relative EUs: {range_eus}\n")
        core_id_list0 = self._create_core_ids_list(group[key_range])
        if len(core_id_list0) > 0:
            core_id_list1 = create_string_from_list(core_id_list0)
            output.write(f"{prefix}Cores in use: {core_id_list1}\n")

    # This protected method returns a spacing symbols for tree-like structures.
    def _get_spacing(self):
        return " " * self._offset

    # This protected method returns a connector symbol for tree-like structures.
    # '└' is used for the last child, and '├' is used for intermediate children.
    def _get_connector(self, is_last):
        if is_last:
            return "└" + ("─" * (self._offset - 2)) + " "
        return "├" + ("─" * (self._offset - 2)) + " "

    # This protected method returns a space symbols for tree-like structures.
    def _get_space(self, is_last):
        if is_last:
            return " " * self._offset
        return "│" + " " * (self._offset - 1)

    # Function to print status as dictionary
    def _print_status(self, output):
        self._print_partition_root(output)
        partition_len = len(self._elements[PARTITIONS])
        root_groups_len = len(self._elements[ROOT_PARTITION_GROUPS])

        main_is_last = not (partition_len > 0 or root_groups_len > 0)

        if len(self._elements[PARTITIONS]) > 0:
            prefix = self._get_space(main_is_last)
            output.write(f"{prefix}\n")
            main_is_last = root_groups_len == 0
            connector = self._get_connector(main_is_last)
            output.write(f"{connector}PARTITIONS:\n")
            for index, (name, partition) in enumerate(self._elements[PARTITIONS].items()):
                is_last = index >= partition_len - 1
                prefix = self._get_space(main_is_last) + self._get_space(False)
                output.write(f"{prefix}\n")
                prefix = self._get_space(main_is_last) + self._get_connector(is_last)
                output.write(f"{prefix}{name}:\n")
                prefix = self._get_space(main_is_last) + self._get_space(is_last)
                self._print_partition(output, partition, prefix)
                groups_nmb = len(partition['groups']) if 'groups' in partition else 0
                if partition['max_groups'] > groups_nmb:
                    output.write(f"{prefix}Maximal number of the groups: "
                                 f"{partition['max_groups']}\n")
                if groups_nmb > 0:
                    output.write(f"{prefix}GROUPS:\n")
                    for group_index, group in enumerate(partition['groups']):
                        is_last = group_index >= groups_nmb - 1
                        self._print_group(output, group, prefix, is_last)

        if len(self._elements[ROOT_PARTITION_GROUPS]) > 0:
            prefix = self._get_space(False)
            output.write(f"{prefix}\n")
            connector = self._get_connector(True)
            prefix = f"{connector}"
            output.write(f"{prefix}ROOT_PARTITION_GROUPS:\n")
            prefix = self._get_spacing()
            for group_index, group in enumerate(self._elements[ROOT_PARTITION_GROUPS]):
                is_last = group_index >= root_groups_len - 1
                self._print_group(output, group, prefix, is_last)

        if not self._errors_in_core_mapping:
            self._print_core_mapping(output)

        self._print_vhca_mapping(output)

    # Main function for work
    def run(self):
        output = self.open_file_for_write()
        if self._options.type == "vhca":
            self._chip_data.read_vhca_data()
            vhca_data = VHCATreeData(self._chip_data.vhca_data)
            if self._options.yaml_format:
                vhca_data.print_dict_as_yaml(output)
            else:
                vhca_data.print_tree(output)
        elif self._options.type == "chip":
            self._chip_data.read_dpa_groups_resources()
            self._chip_data.read_dpa_partitions_resources()
            self._chip_data.read_dpa_resources()
            resource_data = DPAResources(self._chip_data.resource_data)
            if self._options.yaml_format:
                resource_data.write_dict_as_yaml(output)
            else:
                resource_data.print_resources(output)
        elif self._options.type == "resources":
            self._read_and_parse_status()
            if self._options.yaml_format:
                self._print_status_as_yaml(output)
            else:
                self._print_status(output)
        if output != sys.stdout:
            output.close()
