#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Module that contain class GroupHandler

import sys
from ....simenums.simenums import DPADB
from ....exceptions.exception import HandlerException
from ..dpaobjects import DPAObjectsHandler


# Class for create, query and destroy groups
class GroupHandler(DPAObjectsHandler):
    def __init__(self, arg_params, filename):
        super().__init__(arg_params, filename)
        part_id = self._params.partition_id
        self._partition_group = None
        self._partition_group_index = -1
        if part_id:
            found = False
            if not self._partition_list or len(self._partition_list) == 0:
                self._error_message(f"Partition {part_id} does not created yet")
            for index, partition in enumerate(self._partition_list):
                if partition['id'] == part_id:
                    if 'groups' in partition:
                        self._group_list = partition['groups']
                    else:
                        self._group_list = []
                    self._partition_group = partition
                    self._partition_group_index = index
                    found = True
                    break
            if not found:
                self._error_message(f"Partition {part_id} does not created yet")
        else:
            if self._data[DPADB.ROOT_GROUPS]:
                self._group_list = self._data[DPADB.ROOT_GROUPS]
            else:
                self._group_list = []

    def _get_subfunction(self):
        subfunc_dict = {"create": self.__create_group,
                        "query": self.__query_group,
                        "destroy": self.__destroy_group,
                        "info": self.__print_info}
        return subfunc_dict[self._params.subcommand]

    def _set_update(self):
        if self._partition_group:
            self._data[DPADB.PARTITIONS][self._partition_group_index]['groups'] = self._group_list
        else:
            self._data[DPADB.ROOT_GROUPS] = self._group_list
        super()._set_update()

    def _check_for_query(self, group):
        if self._params.name_group:
            return bool(self._params.name_group == group['name'])
        if self._params.group_id:
            return bool(self._params.group_id == group['id'])

        return True

    def __print_info(self):
        print(f"Max number of DPA EU groups: {self.max_eu_groups - 1}")
        print(f"Max number of DPA EUs in one DPA EU group: {self.max_num_dpa_eu_per_group}")
        print(f"Max DPA EU number available to use: {self.max_num_dpa_eu}")
        print("Max EU group name length is 15 chars")

    @property
    def _name(self):
        return "group"

    def __create_group(self):
        if not self._group_list:
            self._group_list = []
        if self._partition_group:
            if self._partition_group['max_num_eu_group'] == len(self._group_list):
                self._error_message(f"Error: There are {len(self._group_list) + 1} EU groups, "
                                    "exceeded Max number of DPA EU groups")
        for group in self._group_list:
            if self._params.name_group == group['name']:
                self._error_message(f"Group name: \"{self._params.name_group}\", is already "
                                    f"taken by group_id: {group['id']}")
        new_group = self.__new_group()
        self._dpa_obj.fill_group(self._params.partition_id, new_group)
        normalized_group = self._dpa_obj.read_group(self._params.partition_id, new_group['id'])
        if not normalized_group:
            raise HandlerException("Cannot create group")
        self._group_list.append(normalized_group)
        print("Group created successfully -")
        self._print_single_elem(normalized_group)
        self._set_update()

    def __query_group(self):
        self._query_elements(self._group_list)

    def _print_single_elem(self, elem):
        print(f"EU group ID: {elem['id']}")
        print(f"EU group name: {elem['name']}")
        if self._params.partition_id:
            print(f"Partition ID: {self._params.partition_id}")
        print(f"Member EUs are: {elem['range']}")

    def __destroy_group_error(self):
        if self._params.name_group:
            print(f"No group with name: \"{self._params.name_group}\" exists.")
        else:
            print(f"Error when querying group id: {self._params.group_id}.")
        sys.exit(1)

    def __destroy_group(self):
        if not self._group_list or len(self._group_list) == 0:
            self.__destroy_group_error()
        for index, group in enumerate(self._group_list):
            name = group['name']
            ident = group['id']
            if self._params.name_group:
                if self._params.name_group == name:
                    self._group_list.pop(index)
                    print(f"Group with name: \"{name}\" (group_id: {ident}) - destructed")
                    self._set_update()
                    return
            else:
                if self._params.group_id == ident:
                    self._group_list.pop(index)
                    print(f"Group with group id: {ident}, was destroyed successfully")
                    self._set_update()
                    return
        self.__destroy_group_error()

    def __new_group(self):
        part_dict = {}
        ident = self._find_free_id(self._group_list)
        part_dict['id'] = ident
        part_dict['range'] = str(self._params.range_eus)
        part_dict['name'] = str(self._params.name_group)
        return part_dict
