#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Module that contain class PartitionHandler

import sys
from ....exceptions.exception import HandlerException
from ....utilities.string import expand_range_from_string
from ..dpaobjects import DPAObjectsHandler


# Class for create, query and destroy partitions.
# Also provide print some information about chip.
class PartitionHandler(DPAObjectsHandler):
    def _get_subfunction(self):
        subfunc_dict = {"info": self.__print_info,
                        "create": self.__create_partition,
                        "query": self.__query_partition,
                        "destroy": self.__destroy_partition}
        return subfunc_dict[self._params.subcommand]

    @property
    def _name(self):
        return "parition"

    def __print_info(self):
        print(f"Max number of DPA EU partitions: {self.max_num_dpa_eu_partitions - 1}")
        print("Max number of VHCAs associated with a single partition: "
              + f"{self.max_num_partition_vhca_id}")
        print(f"Max number of DPA EU groups: {self.max_eu_groups - 1}")
        print("Note- an allocation of a partition consumes from the number of DPA EU "
              "*groups* available to create")
        print(f"Max number of DPA EUs available to use: {self.max_num_dpa_eu}")

    def __create_partition(self):
        part_list = self._partition_list
        if not part_list:
            self._partition_list = []
        new_partition = self.__new_partition()
        self._dpa_obj.fill_partition(new_partition)
        normalized_partition = self._dpa_obj.read_partition(new_partition['id'])
        if not normalized_partition:
            raise HandlerException("Cannot create partition")
        self._partition_list.append(normalized_partition)
        print("Partition created successfully -")
        self._print_single_elem(normalized_partition)
        self._set_update()

    def __query_partition(self):
        self._query_elements(self._partition_list)

    def _print_single_elem(self, elem):
        print(f"EU Partition ID: {elem['id']}")
        print(f"Maximal number of groups: {elem['max_num_eu_group']}")
        vhca_list = elem['vhca_list']
        len_vhca = len(expand_range_from_string(vhca_list))
        print(f"The partition has a total of {len_vhca} associated VHCA IDs, namely: {vhca_list}")
        print(f"Partition's member EUs are: {elem['range']}")

    def __destroy_partition_error(self):
        print(f"Error when trying to destroy EU partition id: {self._params.partition_id}.")
        print("Make sure no EU groups are configured on the partition. Terminating..")
        sys.exit(1)

    def __destroy_partition(self):
        data = self._dpa_obj.get_data()
        if self._params.partition_id not in data['Partitions'].keys():
            self.__destroy_partition_error()
        for index, partition in enumerate(self._partition_list):
            if partition['id'] == self._params.partition_id:
                self._partition_list.pop(index)
                break
        print(f"Partition with partition id: {self._params.partition_id}, "
              "was destroyed successfully")
        self._need_update = True

    def __new_partition(self):
        part_dict = {}
        ident = self._find_free_id(self._partition_list)
        part_dict['id'] = ident
        part_dict['range'] = str(self._params.range_eus)
        part_dict['vhca_list'] = str(self._params.vhca_list)
        part_dict['max_num_eu_group'] = int(self._params.max_num_eu_group)
        part_dict['groups'] = []
        return part_dict
