#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Module that contain SimArgParse

import sys
import argparse


# Class for parsing input arguments for sim-dpaeumgmt
class SimArgParse:
    def __init__(self):
        self._params = None
        self._parser = argparse.ArgumentParser()
        self._subparsers = self._parser.add_subparsers(dest='cmd')
        self.__collect_sub_parsers()

    def __collect_sub_parsers(self):
        partition_subparser = self._subparsers.add_parser("partition",
                                                          description="partions commands.")
        partition_subparser.add_argument('subcommand',
                                         help="subcommand",
                                         choices=["create", "destroy", "query", "info"])
        partition_subparser.add_argument('--dpa_device', '-d',
                                         dest='device', required=True)
        partition_subparser.add_argument('--range_eus', '-r',
                                         dest='range_eus', type=str)
        partition_subparser.add_argument('--vhca_list', '-v',
                                         dest='vhca_list', type=str)
        partition_subparser.add_argument('--max_num_eu_group', '-m',
                                         dest='max_num_eu_group', type=int)
        partition_subparser.add_argument('--id_partition', '-p',
                                         dest='partition_id', type=int)

        group_subparser = self._subparsers.add_parser("eu_group",
                                                      description="group commands.")
        group_subparser.add_argument('subcommand',
                                     help="subcommand",
                                     choices=["create", "destroy", "query", "info"])
        group_subparser.add_argument('--dpa_device', '-d',
                                     dest='device', required=True)
        group_subparser.add_argument('--range_eus', '-r',
                                     dest='range_eus', type=str)
        group_subparser.add_argument('--name_group', '-n',
                                     dest='name_group', type=str)
        group_subparser.add_argument('--id_group', '-g',
                                     dest='group_id', type=int)
        group_subparser.add_argument('--id_partition', '-p', default=0,
                                     dest='partition_id', type=int)

        self._subparsers.add_parser("version", description="version")

        info_subparser = self._subparsers.add_parser("info",
                                                     description="info commands.")
        info_subparser.add_argument('subcommand',
                                    help="subcommand",
                                    choices=["chip", "vhca", "status"])
        info_subparser.add_argument('--dpa_device', '-d',
                                    dest='device', required=True)
        info_subparser.add_argument('--yaml-file', '-y',
                                    dest='yaml_file')

        # Command ext_command contain commands that not used by
        # dpa-resource-mgmt-tool but for working with sim database
        # externally
        ext_subparser = self._subparsers.add_parser("ext_command",
                                                    description="extention commands")
        ext_subparser.add_argument('subcommand',
                                   help="subcommand",
                                   choices=["set_processes", "reset", "init"])
        ext_subparser.add_argument('--dpa_device', '-d',
                                   dest='device', required=True)
        ext_subparser.add_argument('--processes_num', '-p',
                                   dest='processes_num', type=int)

    def parse(self):
        self._params = self._parser.parse_args()
        if not self._params.cmd:
            self._parser.print_help()
            sys.exit(0)

        self.__check_correctness()
        return self._params

    def __no_parameter_error(self, parameter):
        print(f"No parameter \"{parameter}\" in {self._params.cmd} "
              f"{self._params.subcommand}")
        sys.exit(1)

    def __check_partition_correctness(self):
        if self._params.subcommand == "create":
            if not self._params.range_eus:
                self.__no_parameter_error('range_eus')
            if not self._params.vhca_list:
                self.__no_parameter_error('vhca_list')
            if not hasattr(self._params, 'max_num_eu_group'):
                self.__no_parameter_error('max_num_eu_group')
        elif self._params.subcommand == "destroy":
            if not self._params.partition_id:
                self.__no_parameter_error('id_partition')

    def __check_group_correctness(self):
        if self._params.subcommand == "create":
            if not self._params.range_eus:
                self.__no_parameter_error('range_eus')
            if not self._params.name_group:
                self.__no_parameter_error('name_group')
        elif self._params.subcommand == "destroy":
            if not self._params.group_id and not self._params.name_group:
                self.__no_parameter_error('id_group or name_group')

    def __check_ext_command_correctness(self):
        if self._params.subcommand == "set_processes":
            if not self._params.processes_num:
                self.__no_parameter_error('processes_num')

    def __check_correctness(self):
        if self._params.cmd == "partition":
            self.__check_partition_correctness()
        elif self._params.cmd == "eu_group":
            self.__check_group_correctness()
        elif self._params.cmd == "ext_command":
            self.__check_ext_command_correctness()
