#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Module that contain SimArgParse

import sys
import io
from enum import IntEnum
import yaml
from ..exceptions.exception import HandlerException


# Class for read and write simaulator database file
class YAMLSimOperations:
    def __init__(self):
        pass

    # Function to read YAML file and handle potential errors
    def read_database_file(self, filename, arg_data):
        try:
            readed_database = arg_data
            with open(filename, 'r', encoding="utf-8") as file:
                database = yaml.load_all(file, Loader=yaml.SafeLoader)
                key = 0
                for data in database:
                    readed_database[key] = data
                    key += 1
                return readed_database
        except yaml.YAMLError as exc:
            raise HandlerException(f"Error parsing YAML file: {exc}") from exc
        except FileNotFoundError:
            device = filename.split('/')[-1].replace("sim_", "", 1).replace(".yml", "", 1)
            print(f"Database file {filename} is not found.\n"
                  f"Please create it with \"sim-dpaeumgmt ext_command init -d {device}\" command")
            sys.exit(1)

    def open_file_for_write(self, filename):
        if not filename:  # Use stdout by default
            output = sys.stdout
            sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')
        else:
            try:
                # pylint: disable=consider-using-with
                output = open(filename, 'w', encoding="utf-8")
                # pylint: enable=consider-using-with
            # pylint: disable=bare-except
            # pylint: disable=broad-except
            except BaseException:
                print(f"Cannot open {filename} for write")
                sys.exit(1)
            # pylint: enable=broad-except
            # pylint: enable=bare-except
        return output

    def write_dict(self, data, output):
        try:
            out = yaml.dump(data, default_flow_style=False, sort_keys=False)
        except TypeError:
            out = yaml.dump(data, default_flow_style=False)
        output.write(f"{out}")

    def print_dict(self, data, filename):
        output = self.open_file_for_write(filename)
        self.write_dict(data, output)
        if output != sys.stdout:
            output.close()

    def print_chip_dict(self, data, filename):
        keys = ["max_num_dpa_eu_partitions", "max_num_dpa_eu_per_group",
                "max_num_partition_vhca_id", "max_num_dpa_eu"]
        data_to_print = data
        for key in keys:
            del data_to_print[key]
        self.print_dict(data_to_print, filename)

    def print_vhca_dict(self, data, filename):
        keys = ["vhca_id", "parent_vhca_id", "vport_id"]
        str_keys = ["pci_bdf"]
        output = self.open_file_for_write(filename)
        for elem in data:
            prefix = "- "
            for key, value in elem.items():
                if key in keys and value:
                    string_value = f"{value:#x}"
                elif key in str_keys:
                    string_value = f"'{value}'"
                else:
                    string_value = f"{value}"
                output.write(f"{prefix}{key}: {string_value}\n")
                prefix = "  "
        if output != sys.stdout:
            output.close()

    def update_database(self, data, filename):
        output = self.open_file_for_write(filename)
        for key in sorted((k for k in data.keys()
                           if isinstance(k, IntEnum)), key=lambda x: x.value):
            output.write('---\n')
            sub_data = data[key]
            if sub_data:
                self.write_dict(sub_data, output)
        if output != sys.stdout:
            output.close()
