#!/usr/bin/python3
__copyright__ = """
 SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Module for YAML read-write operations

import sys
import os
try:
    import yaml
except ModuleNotFoundError as e:
    sys.exit(f"{e}. Please install it with pip install pyyaml")
from ..exceptions.exception import HandlerException
from ..constants.constants import PARTITIONS, DPA_APPS, VERSION, EU_GROUP_ALLOCATION, \
    NAME, NUM_EUS, RELATIVE_EUS, ROOT_PARTITION_GROUPS, DEFAULT_OUTPUT_YAML_VERSION, \
    AFFINITY_EUS

# Global variable that contain read YAML file name

YAML_FILE_FOR_READ = None

# Inherit class for read YAML files and check for duplicate keys


# pylint: disable=too-many-ancestors
class UniqueKeyLoader(yaml.SafeLoader):
    def construct_mapping(self, node, deep=False):
        mapping = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if key in mapping:
                raise HandlerException(f"Duplicate key {key} found in '{YAML_FILE_FOR_READ}' file")
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping
# pylint: enable=too-many-ancestors

# Class for read and write YAML files


class YAMLOperations:
    def __init__(self):
        self._data = None
        self._database = None
        self._output_version = DEFAULT_OUTPUT_YAML_VERSION
        self._yaml_version = None

    # Function to check and collect data
    def _collect_data(self, database):
        allowed_keys = [PARTITIONS, DPA_APPS, VERSION, ROOT_PARTITION_GROUPS]
        for doc in self._data:
            for key in doc.keys():
                if key not in allowed_keys:
                    raise HandlerException(f"Invalid key - '{key}'")
                if key in database:
                    raise HandlerException(f"Duplicate key - '{key}'")
                database[key] = doc[key]

    # Function for process data
    def process_data(self):
        database = {}
        self._collect_data(database)
        self._database = database

    # Function for get YAML version
    def _get_yaml_version(self):
        for doc in self._data:
            if VERSION in doc.keys():
                self._yaml_version = str(doc[VERSION])
                continue

    # Function for insert partitions if not exist
    def _insert_paritions_if_not_exist(self):
        if PARTITIONS not in self._data[0]:
            print("No partitions found in the YAML file. Adding empty partitions.")
            self._data[0][PARTITIONS] = {}

    # Function to read YAML file and handle potential errors
    def read_yaml_file(self, filename):
        try:
            with open(filename, 'r', encoding="utf-8") as file:
                # pylint: disable=global-statement
                global YAML_FILE_FOR_READ
                YAML_FILE_FOR_READ = filename
                # pylint: enable=global-statement
                self._data = []
                data = yaml.load_all(file, Loader=UniqueKeyLoader)
                for doc in data:
                    if doc:
                        self._data.append(doc)
            self._get_yaml_version()
            if not self._yaml_version:
                raise HandlerException(f"YAML file {filename} must contain version field")
            self._insert_paritions_if_not_exist()
            return self._yaml_version
        except yaml.YAMLError as exc:
            raise HandlerException(f"Error parsing YAML file: {exc}") from exc
        except FileNotFoundError as exc:
            raise HandlerException(f"File not found: {filename} as {exc}") from exc

    # Function to read YAML file and handle potential errors
    def get_yaml_database(self):
        return self._database

    # Function for write groups sub-section in yaml
    def _write_eug(self, outfile, eugdb):
        indent = 0
        outfile.write(' ' * indent + f"{EU_GROUP_ALLOCATION}:\n")
        indent += 2
        for val in eugdb:
            outfile.write(' ' * indent + f"- {NAME}: {val[NAME]}\n")
            outfile.write(' ' * indent + f"  {NUM_EUS}: {val[NUM_EUS]}\n")
        outfile.write("\n")

    # Function for write app sub-section in yaml
    def _write_apps_group(self, outfile, app, indent):
        number_groups = len(app[EU_GROUP_ALLOCATION])
        if not number_groups:
            return
        groups = app[EU_GROUP_ALLOCATION]
        outfile.write(' ' * indent + f"number_of_{EU_GROUP_ALLOCATION}: {number_groups}\n")
        commentstring = ' ' * indent + '#'
        name_list = []
        id_list = []
        for key, item in groups.items():
            (_, name) = key
            name_list.append(name)
            id_list.append(item['ID'])
        commentstring = ' ' * indent + '# ' + ", ".join(name_list)
        outfile.write(f"{commentstring}\n")
        outfile.write(' ' * indent + f"{EU_GROUP_ALLOCATION}: ")
        outstring = ", ".join(id_list)
        outfile.write(f"[{outstring}]\n")

    # Function for write groups section in yaml
    def _write_apps_eu(self, outfile, app, indent):
        if len(app[RELATIVE_EUS]) == 0:
            return

        abs_apps = sorted(app[AFFINITY_EUS])
        outfile.write(' ' * indent + f"# absolute EUs: {abs_apps}\n")
        arr_apps = sorted(app[RELATIVE_EUS])
        outfile.write(' ' * indent + f"number_of_{AFFINITY_EUS}: "
                      + f"{len(arr_apps)}\n")
        outfile.write(' ' * indent + f"{AFFINITY_EUS}: {arr_apps}\n")

    # Function for write application section in yaml
    def _write_apps(self, outfile, appdb):
        indent = 0
        outfile.write(' ' * indent + f"{DPA_APPS}:\n")
        indent += 2
        for app in appdb:
            outfile.write(' ' * indent + f"- {NAME}: {app[NAME]}\n")
            indent += 2
            if VERSION in app:
                outfile.write(' ' * indent + f"{VERSION}: {app[VERSION]}\n")
            self._write_apps_group(outfile, app, indent)
            self._write_apps_eu(outfile, app, indent)
            indent -= 2

    # Function for write single yaml file
    def write_yaml_file(self, key, yaml_db, output):
        filename = os.path.join(output, str(key) + '.yaml')
        with open(filename, 'w', encoding="utf-8") as outfile:
            outfile.write(f"version: {self._output_version}\n")
            self._write_apps(outfile, yaml_db[DPA_APPS])

    # Function for write dictionary as yaml
    def write_dict(self, data, output):
        try:
            out = yaml.dump(data, default_flow_style=False, sort_keys=False)
        except TypeError:
            out = yaml.dump(data, default_flow_style=False)
        output.write(f"{out}")

    # Function for read yaml from string
    @staticmethod
    def get_object_from_yaml(yaml_strings):
        obj = yaml.safe_load(yaml_strings)
        return obj

    # Function for set output version (if needed)
    def output_version(self, version):
        self._output_version = version
